#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <ctime>
#include <iostream>
#include <vector>

using namespace std;

double dx[] = {1, 0, -1, 0};
double dy[] = {0, -1, 0, 1};


const int START_DIR = 0;
const double SIZE = 1;
const double THRESHOLD = 0.01;
double EPS = 0.000000000001;

const int ITERATIONS = 10000;

double PI = acos(-1.0);

World *world = 0;
CPlayer *player;
int dir = START_DIR;

double range(double x, double y)
{
	return sqrt(x * x + y * y);
}

double dist(double x1, double y1, double x2, double y2)
{
	return sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2));
}

double dot_product(double x1, double y1, double x2, double y2)
{
	return x1 * x2 + y1 * y2;
}

bool eq(double a, double b)
{
	return fabs(a - b) < EPS;
}




vector<int> get_alive_monsters()
{
	vector<int> alive;
	for ( int j = 0; j < world->MAXMONSTERS; j++)
		if (world->monsters[j].ClassName >= 0) {
			alive.push_back(j);
		}
	return alive;
}

void shoot_monster(int m)
{
	world->shoot(world->monsters[m].PositionX - world->player.PositionX,
			   world->monsters[m].PositionY - world->player.PositionY);
}

bool shooters_over_threshold()
{
	vector<int> alive = get_alive_monsters();

	int shooter = 0;
	for (int i = 0; i < alive.size(); i++)
		if(world->monsters[alive[i]].BulletDamage > 0)
			shooter++;

	return double(shooter) / alive.size() > THRESHOLD; 
}

void shoot_random_monster()
{
	vector<int> alive = get_alive_monsters();

	if (alive.empty())
		return;

	int r = rand() % alive.size();
	world->shoot(world->monsters[r].PositionX - world->player.PositionX,
			   world->monsters[r].PositionY - world->player.PositionY);
}


void shoot_first_monster(){
		for ( int j = 0; j < world->MAXMONSTERS; j++)
			if (world->monsters[j].ClassName >= 0) {
				//Shoot the bastard! (Shoot in the direction of the monster)
				world->shoot(world->monsters[j].PositionX - world->player.PositionX,
						   world->monsters[j].PositionY - world->player.PositionY);
				break;
			}
}

void shoot_weakest_monster()
{
	vector<int> alive = get_alive_monsters();
	if (alive.empty())
		return;

	int weakest = 0;
	for (int i = 1;i < alive.size();i++)
	{
		if (world->monsters[alive[i]].HitPoints < world->monsters[alive[weakest]].HitPoints)
			weakest = i;
	}


	shoot_monster(alive[weakest]);
}

int get_nearest_monster(vector<int> monsters)
{
	if (monsters.empty())
		return -1;
	int nearest = 0;
	for (int i = 1;i < monsters.size();i++)
	{
		double mx, my, nx, ny;
		mx = world->monsters[monsters[i]].PositionX;
		my = world->monsters[monsters[i]].PositionY;

		nx = world->monsters[monsters[nearest]].PositionX;
		ny = world->monsters[monsters[nearest]].PositionY;

		if (dist(mx, my, player->PositionX, player->PositionY) < dist(nx, ny, player->PositionX, player->PositionY))
		{
			nearest = i;
		}
	}
	return monsters[nearest];
}

void shoot_nearest_monster()
{
	vector<int> alive = get_alive_monsters();
	if (alive.empty())
		return;

	shoot_monster(get_nearest_monster(alive));
}

int getm_on_dir()
{
	vector<int> alive = get_alive_monsters();
	if (alive.empty())
		return -1;


	double cos_angle = cos(PI  / (180.0 / 10.0));

	vector<int> monsters_in_angle;
	for (int i = 0; i < alive.size();i++)
	{
		double mx, my;
		
		mx = world->monsters[alive[i]].PositionX - player->PositionX;
		my = world->monsters[alive[i]].PositionY - player->PositionY;

		double cosa = dot_product(dx[dir], dy[dir], mx, my) / range(mx, my);
		if(cosa > cos_angle)
		{
			monsters_in_angle.push_back(alive[i]);
		}
	}

	int near = get_nearest_monster(monsters_in_angle);

	return near;
}

void shoot_at_center()
{
	double sx = 0.5 - world->player.PositionX;
	double sy = 0.5 - world->player.PositionY;
	world->shoot(sx, sy);
}


bool hit(int monster, double a)
{
	double bulletVel = world->player.BulletVelocity;

	double v1x = cos(a) * bulletVel;
	double v1y = sin(a) * bulletVel;

	double v2x = world->monsters[monster].VelocityX;
	double v2y = world->monsters[monster].VelocityY;

	double bx = world->player.PositionX;
	double by = world->player.PositionY;

	double mx = world->monsters[monster].PositionX;
	double my = world->monsters[monster].PositionY;

	/*
	double mr = range(mx, my);
	mx /= mr;
	my /= mr;*/
	
	//double t = (my * bx - mx * by) / (vy * mx - my * vx);

	double t = (bx - mx) / (v2x - v1x); 

	if(t > 0)
	{
		double x1 = bx + v1x * t;
		double y1 = by + v1y * t;

		double x2 = mx + v2x * t;
		double y2 = my + v2y * t;

		if(range(x2 - x1, y2 - y1) <= world->monsters[monster].Radius / 2)
		{
			if(x1 > 0 && y1 > 0 && x1 <= 1 && y1 <= 1)
			{
				/*
				cout << "BULLET START: " << bx << ' ' << by << endl;
				cout << "BULLET VECTOR: " << vx << ' ' << vy << endl;
				cout << "MONSTER START: " << world->monsters[monster].PositionX << ' ' << world->monsters[monster].PositionY << endl;
				cout << "MONSTER VECTOR: " << mx << ' ' << my << endl;
				cout << "TIME: " << t << endl;
				cout << "DIST: " << range(x2 - x1, y2 - y1) << endl;
*/
				return true;
			}

			
		}
	}
	return false;
}

bool try_shoot_accurate(int monster, double & ret)
{
	double da = 2 * PI / ITERATIONS;

	for(double a = 0; a < 2 * PI; a += da)
		if(hit(monster, a))
		{
			ret = a;
			return true;
		}
	return false;
}

void shoot_at_gunner()
{
	vector<int> alive = get_alive_monsters();
	vector<int> gunners;

	for (int i = 0; i < alive.size(); i++)
		if(world->monsters[alive[i]].BulletDamage > 0)
			gunners.push_back(alive[i]);

	int shooter = -1;
	double a = -1;
	for(int it = 0; it < 10; it++)
	{
		int now = rand() % gunners.size();
		if(try_shoot_accurate(gunners[now], a))
		{		
			//cout << "SHOOT AT: " << aliv[i] << ", " << a << endl;
			world->shoot(cos(a), sin(a));
			break;
		}		
	}
}

void shoot()
{
	if(!shooters_over_threshold())
	{
		int ret = getm_on_dir();
		if(ret != -1)
		{	
			cout << "ONDIR\n";
			world->shoot(world->monsters[ret].PositionX - world->player.PositionX, world->monsters[ret].PositionY - world->player.PositionY);
		}
		else
		{
			//shoot_random_monster();
			//shoot_weakest_monster();
			shoot_nearest_monster();
		}
	}
	else
	{
		shoot_at_gunner();		
	}
}


bool can_move(double x, double y)
{
	return true;
}

int main(int argc, char** argv) {
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	world = new World(argv[1], atoi(argv[2]));

	player = &world->player;
	float alpha = 0.0;	
	
	int shoot_dir = START_DIR;
	double prev_x = -1;
	double prev_y = -1;
	while (true) {
		if (eq(world->player.PositionX, prev_x) && eq(player->PositionY, prev_y))
			dir = (dir + 1) % 4;			

		prev_x = world->player.PositionX;
		prev_y = world->player.PositionY;

		cout <<world->player.PositionX << " " <<world->player.PositionY <<" " << dir<<"\n";


		shoot();

		world->run(dx[dir], dy[dir]);
		world->makeMove();
	}

	return 0;
}

		//world->makeMove();

		//printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
		//		world->player.Time, world->player.HitPoints,
		//		world->player.PositionX, world->player.PositionY);

		//alpha += 0.02;

		//world->run(cos(alpha), sin(alpha));

		////Choose the first monster that is still alive
		//for ( int j = 0; j < world->MAXMONSTERS; j++)
		//	if (world->monsters[j].ClassName >= 0) {
		//		//Shoot the bastard! (Shoot in the direction of the monster)
		//		world->shoot(world->monsters[j].PositionX - world->player.PositionX,
		//				   world->monsters[j].PositionY - world->player.PositionY);
		//		break;
		//	}

//int player_on_edge()
//{
//	double fx = fabs(world->player.PositionX);
//	double fy = fabs(world->player.PositionY);
//
//	if (eq(fx, 0))
//		return 0;
//
//	if (eq(fx, SIZE))
//		return 1;
//
//	if (eq(fy, 0))
//		return 2;
//
//	if (eq(fy, SIZE))
//		return 3;
//	return -1;
//}