#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#define PI 3.1416927

double rat (World *world,int n) {
	return pow((double)abs(world->monsters[n].PositionX - world->player.PositionX),2) 
		+ pow((double)abs(world->monsters[n].PositionY - world->player.PositionY),2);
}

#define pos_count  16
#define pos_dist 0.1
#define cell_count 20
#define TURNS 10

double cell_pow[cell_count][cell_count];
double pos_pow[pos_count];

double min_ (double a, double b) {
	if (a > b) return b;
	return a;
}
void fill (World *world) {
	for (int i = 0; i < world->MAXMONSTERS; i ++) {
		if (world->monsters[i].ClassName >= 0) {
			double ax = world->monsters[i].PositionX;
			double ay = world->monsters[i].PositionY;
			double Vx = world->monsters[i].VelocityX;
			double Vy = world->monsters[i].VelocityY;
			cell_pow[(unsigned)(world->monsters[i].PositionX * cell_count)][(unsigned)(world->monsters[i].PositionY * cell_count)] +=
				world->monsters[i].BiteDamage;
			for (int j = 0; j < cell_count; j ++) {
				for (int k = 0; k < cell_count; k ++) {
					double x = j / cell_count;
					double y = k / cell_count;
					cell_pow[j][k] += world->monsters[i].BiteDamage * sqrt(Vx * Vx + Vy * Vy - ((Vx * (y - ay) - (x - ax) * Vy) / sqrt(pow(ax - x,2) + pow((ay - y),2)))) / (sqrt(pow(ax - x,2) + pow((ay - y),2)));
				}
			}

		}
	}
}

int main(int argc, char** argv) {
	int g = -1;
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));

	//Infinite loop
		int xx = 0;
		int yy = 0;
		double x,y;
	double dx = 0.01,dy = 0.01;
	while (true) {
		g++;
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf ",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);
		int min = 0;
		/*fill (world);
		for (int i = 0; i < cell_count; i ++) {
			for (int j = 0; j < cell_count; j ++) {
				if (cell_pow[i][j] < cell_pow[xx][yy]) {
					xx = i;
					yy = j;
				}
			}
		}
		x = xx / cell_count;
		y = yy / cell_count;
		world->run(x,y);*/
		for (int i = 0; i < pos_count; i ++) {
			pos_pow[i] = 0;
			double alpha = (2 * PI * i) / pos_count;
			//double x = world->player.PositionX + sin(alpha);
			//double y = world->player.PositionY + cos(alpha);
			double x = sin(alpha) * pos_dist;
			double y = cos(alpha) * pos_dist;
			for (int j = 0; j < world->MAXMONSTERS; j ++) {
				if (world->monsters[j].ClassName >= 0) {
					double Vx = world->monsters[j].VelocityX;
					double Vy = world->monsters[j].VelocityY;
					pos_pow[i] += (x * x + y * y) * 
						world->monsters[j].BiteDamage;
				}
			}
			for (int j = 0; j < world->MAXMISSILES; j ++) {
				if (world->missiles[j].ClassName >= 0) {
					if(abs(world->missiles[j].PositionX - world->player.PositionX - x) <= 0.1 && 
						abs(world->missiles[j].PositionY - world->player.PositionY - y) <= 0.1) {
						pos_pow[i] += pow(world->missiles[j].Damage,3);
					}
				}
			}
			x += world->player.PositionX;
			y += world->player.PositionY;
			//if (!(x <= 0 || x >= 1 || y <= 0 || y >= 1)) 
			if (pos_pow[min] > pos_pow[i]) min = i;
		}
		double alpha = (2 * PI * min) / pos_count;
		double x = sin(alpha) * pos_dist;
		double y = cos(alpha) * pos_dist;
		printf("%.4lf\n",pos_pow[min]);
		//if (!(g % TURNS)) world->run(x,y);
		x = 100 / (rand() % 10 + 1);
		if (rand() % 2) x *= -1;
		y = 100 / (rand() % 10 + 1);
		if (rand() % 2) y *= -1;
		//if (!(rand() % 10)) world->run(x,y);
		x = world->player.PositionX;
		y = world->player.PositionY;
		if (abs(x - dx) <= 0.01 && abs(y - dy) <= 0.01) {
			if (dx == 0.01 && dy == 0.01) dy = 1 - dy; else
			if (dx == 0.01 && dy == 0.99) dx = 1 - dx; else
			if (dx == 0.99 && dy == 0.99) dy = 1 - dy; else dx = 1 - dx;
		}
		world->run(dx - x,dy - y);

		double max_k = -1;
		int ind = 0;
		for (int i = 0; i < world->MAXMONSTERS; i ++) {
			if (world->monsters[i].ClassName >= 0) {
				int t = 0;
				double ax = world->monsters[i].PositionX;
				double ay = world->monsters[i].PositionY;
				for (int j = 0; j < world->MAXMONSTERS; j ++) {
					double bx = world->monsters[i].PositionX;
					double by = world->monsters[i].PositionY;
					double h = abs((bx - ax) * (by - y) - (bx - x) * (by - ay)) / 
						sqrt(pow(ax - x, 2) + pow(ay - y,2));
					if (h < world->monsters[j].Radius) {
						t = 1;
						break;
					}
				}
				if (!t) {
					x -= ax;
					y -= ay;
					double Vx = world->monsters[i].VelocityX;
					double Vy = world->monsters[i].VelocityY;
					double k = sqrt(world->monsters[i].Radius)
						/ ((Vx * y - Vy * x) *  world->monsters[i].HitPoints * (x * x + y * y));	
					if (max_k < k) {
						max_k = k;
						ind = i;
					}
				}
			}
		}
		world->shoot(world->monsters[ind].PositionX - world->player.PositionX,
					 world->monsters[ind].PositionY - world->player.PositionY);	
		//Run in circles (set direction)
		//world->run(cos(alpha), sin(alpha));
		/*for ( int j = 0; j < world->MAXMONSTERS; j++)
			if (world->monsters[j].ClassName >= 0) {
				//Shoot the bastard! (Shoot in the direction of the monster)
				world->shoot(world->monsters[j].PositionX - world->player.PositionX,
						   world->monsters[j].PositionY - world->player.PositionY);
				break;
			}*/
		int j;
		do {
			j = rand() % world->MAXMONSTERS;
		}while (world->monsters[j].ClassName < 0);
			world->shoot(world->monsters[j].PositionX - world->player.PositionX,
						   world->monsters[j].PositionY - world->player.PositionY);

	}

	return 0;
}
