#define _CRT_SECURE_NO_WARNINGS
#pragma comment (linker, "/STACK:16777216")
#include <iostream>
#include <algorithm>
#include <numeric>
#include <vector>
#include <queue>
#include <list>
#include <stack>
#include <string>
#include <fstream>
#include <math.h>
#include <limits>
#include <set>
#include <map>
#include <sstream>
#include <stdio.h>
#include <time.h>
#include <memory.h>
#include <cassert>
using namespace std;

#include "world.h"
#include <cstdio>
#include <cstring>

///////////////// macros and typedefs ///////////////////
#define rep(i, n) for (int i = 0, _n = (n); i < _n; ++i)
#define repd(i, n) for (int i = (n)-1; i >= 0; --i)
#define DEB(k) cerr<<"debug: "#k<<"="<<k<<endl;
#define CLEAR(a) memset((a), 0, sizeof(a));
#define all(c) (c).begin(), (c).end()
#define mp(a, b) make_pair(a, b)
#define l(c) (int)((c).size())
#define sqr(a) ((a)*(a))
#define inf 0x7f7f7f7f
#define pb push_back
#define ppb pop_back
#define x first
#define y second
typedef long long ll;
typedef vector<int> vi;
typedef pair<int,int> pi;

#define eps (1e-14)
const double Pi = 2*acos(0.0);

struct pt {
   double x, y;
   pt() : x(0), y(0) {}
   pt(double X, double Y) : x(X), y(Y) {}
   double len() { return sqrt(len2()); }
   double len2() { return x*x+y*y; }
};

inline bool eq(double a, double b) { return fabs(a-b) < eps; }
inline bool lo(double a, double b) { return a+eps < b; }
inline double deg(double ang) { return ang*180/Pi; }
inline double rad(double ang) { return ang*Pi/180; }

pt operator +(pt a, pt b) { return pt(a.x+b.x, a.y+b.y); }
pt operator -(pt a, pt b) { return pt(a.x-b.x, a.y-b.y); }
pt operator *(pt a, double k) { return pt(k*a.x, k*a.y); }
double operator *(pt a, pt b) { return a.x*b.x + a.y*b.y; }
double operator ^(pt a, pt b) { return a.x*b.y - a.y*b.x; }
bool operator <(pt a, pt b) { return lo(a.x, b.x) || eq(a.x, b.x) && lo(a.y, b.y); }
bool operator ==(pt a, pt b) { return eq(a.x, b.x) && eq(a.y, b.y); }

pt rotate(pt p, double alpha, pt c) {
   double acos = cos(alpha), asin = sin(alpha);
   return pt((p.x-c.x)*acos - (p.y-c.y)*asin + c.x,
             (p.x-c.x)*asin + (p.y-c.y)*acos + c.y);
}

// Get polar angle in the range [0, 2Pi) or -1 for (0, 0)
double polarAngle(pt a) {
   if (eq(a.x, 0) && eq(a.y, 0)) return -1;
   double p = atan2(a.y, a.x);
   return p < 0 ? p+2*Pi : p;
}
/////////////////////////////////////////////////////////

World* world;
double R = 0.47;
double LOOK = Pi/7;
vi alive;
bool start;

double distToMob(int k) {
   return sqrt(sqr(world->monsters[k].PositionX - world->player.PositionX) +
               sqr(world->monsters[k].PositionY - world->player.PositionY));
}

void printInfo() {
   pt p = pt(world->player.PositionX, world->player.PositionY);
   pt c = pt(0.5, 0.5);
   fprintf(stderr, "frame : %d, HP: %.2lf x: %.2lf y:%.2lf (%.2lf %.2lf Len: %.2lf)\n",
		world->player.Time, world->player.HitPoints,
		world->player.PositionX, world->player.PositionY,
      world->player.VelocityX, world->player.VelocityY,
      (p-c).len());   
}

vi aliveMonsters() {
   vi res;
   for (int i = 0; i < world->MAXMONSTERS; ++i) {
      if (world->monsters[i].ClassName >= 0) {
         res.push_back(i);
      }
   }
   return res;
}
//
//vi aliveMonsters() {
//   vi res;
//
//   bool willKill[world->MAXMONSTERS] = {0};
//
//   for (int i = 0; i < world->MAXMISSILES; ++i) {
//      if (world->missiles[i].ClassName >= 0) {
//         
//      }
//   }
//
//   for (int i = 0; i < world->MAXMONSTERS; ++i) {
//      if (world->monsters[i].ClassName >= 0) {
//         res.push_back(i);
//      }
//   }
//   return res;
//}


pt reachOrbit() {
   pt v = pt(world->player.VelocityX, world->player.VelocityY);
   pt c = pt(0.5, 0.5);
   pt p = pt(world->player.PositionX, world->player.PositionY);
   if (v.len() < eps) v = pt(0, 1);
   else v = p-c;
   return v;
}

pt nextOrbit() {
   pt p = pt(world->player.PositionX, world->player.PositionY);
   pt c = pt(0.5, 0.5);
   if (p.y > 0.5+R) {
      if (p.x < 0.5+R) return pt(1, 0);
      else return pt(0, -1);
   }
   else if (p.x > 0.5+R) {
      if (p.y > 0.5-R) return pt(0, -1);
      else return pt(-1, 0);
   }
   else if (p.y < 0.5-R) {
      if (p.x < 0.5-R) return pt(0, 1);
      else return pt(-1, 0);
   }
   else {
      return pt(0, 1);
   }
   /*double ang = polarAngle(p-c);
   double len = (p-c).len();
   ang -= 0.02;
   pt ret = c + pt(len*cos(ang), len*sin(ang));
   return (ret-p);*/
}

void setMove() {
   pt p = pt(world->player.PositionX, world->player.PositionY);
   pt c = pt(0.5, 0.5);
   pt nxt;
   if (start || (p-c).len() < R)
      nxt = reachOrbit(), start = false;
   else
      nxt = nextOrbit();
	world->run(nxt.x, nxt.y);
}

double norm(double ang) {
   while (ang < 0) ang += 2*Pi;
   while (ang > 2*Pi) ang -= 2*Pi;
   return ang;
}

bool inside(double alpha, double beta) {
   alpha = norm(alpha);
   beta = norm(beta);
   double ang = norm(alpha-beta);
   return (ang < LOOK);
}

void setShoot() {
   alive = aliveMonsters();
   int m = l(alive);
   pt p = pt(world->player.PositionX, world->player.PositionY);
   pt v = pt(world->player.VelocityX, world->player.VelocityY);
   double ang_v = polarAngle(v);
   int mini = -1;
   double dist = inf;
   rep (i, m) {
      pt mob = pt(world->monsters[alive[i]].PositionX, world->monsters[alive[i]].PositionY);
      double ang_m = polarAngle(mob-p);
      if (inside(ang_v, ang_m)/* || distToMob(alive[i]) < MIN_DIST*/) {
         double dist2 = (mob-p).len();
         if (dist2 < dist)
            dist = dist2, mini = i;
      }
   }
   if (mini == -1 || dist > 0.12) {
      mini = rand()%m;
      /*rep (i, m) {
         pt mob = pt(world->monsters[alive[i]].PositionX, world->monsters[alive[i]].PositionY);
         double dist2 = (mob-p).len();
         if (dist2 < dist)
            dist = dist2, mini = i;
      }*/
   }
   pt ret = pt(world->monsters[alive[mini]].PositionX, world->monsters[alive[mini]].PositionY);
   ret = ret - p; 
   world->shoot(ret.x, ret.y);
}

void step() 
{
   setMove();
   setShoot();
   world->makeMove();
   printInfo();
}

void initialize(char** a) {
   srand(time(0));
   world = new World(a[1], atoi(a[2]));
   start = true;
}

int main(int argc, char** argv)
{
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}
   initialize(argv);
	while (true) step();
	return 0;
}
