#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <iostream>
#include <cstdlib>
using namespace std;

World *world;

const double pi = 3.14159265358979;
const double eps=1e-5;

template <class T>
inline T sqr(const T &a) { return a*a; }

const int SQ_CNT=7,SQ_AROUND=2;
const double FIND_DIST = 0.2;
const int FIND_CNT = 30;
const int FIND_DIST_CNT = 5;
const int VERY_DANGER = 1000;
const int AIM_ITERATIONS = 10;
const double DANGER_LEN = sqr(0.07);
const double VERY_DANGER_LEN = sqr(0.03);
const int PROB_SHOOT=70;
const int FEW_MONSTERS=30;
const double DANGER_ANGLE = pi/12;


struct point {
	double x, y;
	point(double x = 0.0, double y = 0.0):x(x), y(y) {}
	point operator =(const point &a) {
		this->x = a.x;
		this->y = a.y;
		return *this;
	}
};

double norm(const point &a) {
	return a.x*a.x + a.y*a.y;
}
double len(const point &a) {
	return sqrt(norm(a));
}
point operator -(const point &a, const point &b) {
	return point(a.x-b.x, a.y-b.y);
}
point operator +(const point &a, const point &b) {
	return point(a.x+b.x, a.y+b.y);
}

point operator *(double a, const point &b) {
	return point(a*b.x, a*b.y);
}

double operator^(point a,point b) {
	return a.x*b.x+a.y*b.y;
}

double operator*(point a,point b) {
	return a.x*b.y-a.y*b.x;
}

double ang(point a,point b,point c) {
	b=b-a;
	c=c-a;
	return fabs(atan2(b*c,b^c));
}

inline bool is_valid(const point &a) {
	if (a.x < eps) return false;
	if (a.y < eps) return false;
	if (a.x > 1-eps) return false;
	if (a.y > 1-eps) return false;
	return true;
}

inline double area_pow(double a) { return sqr(a*a); }
inline bool in_true_area(const point &a) {
	return area_pow(a.x - 0.5) + area_pow(a.y - 0.5) < area_pow(0.5) + eps;
}

void move();
void shoot();
point aim(int i);
double danger(point);


inline point monster_pos(int id) {
	return point(world->monsters[id].PositionX, world->monsters[id].PositionY);
}

inline point player_pos() {
	return point(world->player.PositionX, world->player.PositionY);
}

inline point missile_pos(int id) {
	return point(world->missiles[id].PositionX, world->missiles[id].PositionY);
}

point where_monster_after(int id, double dt) {
	double mv = world->monsters[id].MaxVelocity;
	point v = point (world->monsters[id].VelocityX, world->monsters[id].VelocityY);
	return monster_pos(id) + dt*v;
}

inline double dist_to_monster(int i) {
	double mx=world->monsters[i].PositionX;
	double my=world->monsters[i].PositionY;
	double px=world->player.PositionX;
	double py=world->player.PositionY;
	return sqrt((mx-px)*(mx-px)+(my-py)*(my-py));
}

inline point speed_aim(int i) {
	if (i==-1) return point(0,0);
	double mx=world->monsters[i].PositionX;
	double my=world->monsters[i].PositionY;
	double px=world->player.PositionX;
	double py=world->player.PositionY;
	double vx=world->monsters[i].VelocityX;
	double vy=world->monsters[i].VelocityY;
	double t=dist_to_monster(i)/world->player.BulletVelocity;
	mx+=t*vx;
	my+=t*vy;
	return point(mx - px, my - py);
}

inline point accel_aim(int i) {
	if (i==-1) return point(0,0);
	double mx=world->monsters[i].PositionX;
	double my=world->monsters[i].PositionY;
	double px=world->player.PositionX;
	double py=world->player.PositionY;
	double vx=world->monsters[i].VelocityX;
	double vy=world->monsters[i].VelocityY;
	point f(mx,my),p=player_pos();
	for (int j=0;j<AIM_ITERATIONS;j++) {
		double t=sqrt(norm(f-p))/world->player.BulletVelocity;
		f=where_monster_after(i,t);
	}
	return point(f.x - px, f.y - py);
}

void aim_stupid(int i) {
	double mx=world->monsters[i].PositionX;
	double my=world->monsters[i].PositionY;
	double px=world->player.PositionX;
	double py=world->player.PositionY;
	world->shoot(mx - px, my - py);
}

int select_monster() {
	int mx=-1;
	for ( int j = 0; j < world->MAXMONSTERS; j++)
		if (world->monsters[j].ClassName >= 0) {
			if (mx==-1 || dist_to_monster(mx)>dist_to_monster(j)) mx=j;
	}
	return mx;
}

int monster_count()
{
	int res=0;
	for ( int j = 0; j < world->MAXMONSTERS; j++) 
		if (world->monsters[j].ClassName >= 0) 
			res++;
	return res;
}

int smart_select_monster() {
	int mc=monster_count();
	if (mc==0) return -1;
	if (mc<FEW_MONSTERS) {
		int ost=rand()%mc;
		for ( int j = 0; j < world->MAXMONSTERS; j++) {
			if (world->monsters[j].ClassName >= 0) {
				ost--;
				if (ost<0) {
					return j;
				}
			}
		}

	}
	int mx=-1;
	for ( int j = 0; j < world->MAXMONSTERS; j++) {
//			cerr << world->monsters[j].BulletDamage << endl;
		if (world->monsters[j].ClassName >= 0 && world->monsters[j].BulletVelocity>eps) {
			if (mx==-1 || dist_to_monster(mx)>dist_to_monster(j)) {
				mx=j;
			}
		}
	}
/*	if (mx!=-1) return mx;
	for ( int j = 0; j < world->MAXMONSTERS; j++)
		if (world->monsters[j].ClassName >= 0) {
			if (world->monsters[j].HitPoints<1-eps) {
				if (mx==-1 || dist_to_monster(mx)>dist_to_monster(j))
					mx=j;
			}
	}*/
	if (mx!=-1 && (rand()%100)<PROB_SHOOT) return mx;
	for ( int j = 0; j < world->MAXMONSTERS; j++)
		if (world->monsters[j].ClassName >= 0) {
			if (mx==-1 || dist_to_monster(mx)>dist_to_monster(j)) mx=j;
	}
	return mx;
}

void main_shoot() {
	//      
	//    
	//    
	//    
	//    ?
	int to=smart_select_monster();
	if (to==-1) return;
	point p = aim(to);
	world->shoot(p.x,p.y);
}

int squares[SQ_CNT][SQ_CNT];

void sq_review() {
	memset(squares,0,sizeof squares);
	for ( int j = 0; j < world->MAXMONSTERS; j++)
		if (world->monsters[j].ClassName >= 0) {
			double mx=world->monsters[j].PositionX;
			double my=world->monsters[j].PositionY;
			squares[(int)(SQ_CNT*mx-eps)][(int)(SQ_CNT*my-eps)]++;
		}
}

double usual_danger(point x) {
	double retval = 0;
	for ( int j = 0; j < world->MAXMONSTERS; j++)
		if (world->monsters[j].ClassName >= 0) {
			retval += 1 / (norm(x - monster_pos(j)) + eps);
		}
	for ( int j = 0; j < world->MAXMISSILES; j++)
		if (world->missiles[j].ClassName >= 0) {
			retval += 1 / (norm(x - missile_pos(j)) + eps);
		}
	return retval;
}

double smart_danger(point x) {
	double retval = 0;
	for ( int j = 0; j < world->MAXMONSTERS; j++)
		if (world->monsters[j].ClassName >= 0) {
			retval += (world->monsters[j].BulletDamage + world->monsters[j].BiteDamage) / (norm(x - monster_pos(j)) + eps);
		}
/*	for ( int j = 0; j < world->MAXMISSILES; j++)
		if (world->missiles[j].ClassName >= 0) {
			retval += world->missiles[j].Damage / (norm(x - missile_pos(j)) + eps);
		}*/
	return retval;
}

void main_move() {
	//    
	//  ,   
	//   
//	sq_review();
/*	double px=world->player.PositionX;
	double py=world->player.PositionY;
	int psx=(int)(SQ_CNT*px-eps);
	int psy=(int)(SQ_CNT*py-eps);
	int qi=0,qj=0;
	for (int i=psx-SQ_AROUND;i<=psx+SQ_AROUND;i++)
		for (int j=psy-SQ_AROUND;j<=psy+SQ_AROUND;j++) {
			if (i<0 || i>=SQ_CNT) break;
			if (j<0 || j>=SQ_CNT) break;
			if (squares[i][j]<squares[qi][qj]){
				qi=i;qj=j;
			}
		}
	double qx=(qi+1./2)/SQ_CNT;
	double qy=(qj+1./2)/SQ_CNT;
	world->run(qx-px,qy-py);*/
	point best = player_pos();
	double best_danger = danger(best);
	for (int i = 0; i < FIND_CNT; i++) {
		double alpha = 2*pi*i/FIND_CNT;
		for (int j = 1; j <= FIND_DIST_CNT; j++) {
			point cur = player_pos() + ((double)j/FIND_DIST_CNT*FIND_DIST)*point(cos(alpha), sin(alpha));
			if (!in_true_area(cur)) continue;
			double cur_danger = danger(cur);
			if (cur_danger < best_danger) {
				best_danger = cur_danger;
				best = cur;
			}
		}
	}
	point res = best - player_pos();
	world->run(res.x, res.y);
	for ( int j = 0; j < world->MAXMONSTERS; j++)
		if (world->monsters[j].ClassName >= 0) {
			if (ang(player_pos(),best,monster_pos(j))<DANGER_ANGLE && norm(monster_pos(j)-player_pos())<DANGER_LEN || norm(monster_pos(j)-player_pos())<VERY_DANGER_LEN) {
				point p=aim(j);
				world->shoot(p.x,p.y);
			}
		}
}

point aim(int i) { return accel_aim(i); }
void move() { main_move(); }
void shoot() { main_shoot(); }
double danger(point x) { return usual_danger(x); }


int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	world = new World(argv[1], atoi(argv[2]));
	//Create the world data (and establish connection with the server)
	float alpha = 0.0;

	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		/*printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf danger:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY, danger(player_pos()));*/
		if (world->player.HitPoints < -eps) break;
		shoot();move();


	}

	return 0;
}
