#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <algorithm>
#include <queue>
//#include <iostream>
#pragma comment(linker,"/STACK:10000000")
using namespace std;

const int grid=13; //       
const float inf=1e7f;
const float eps=1.0f;
int dx[4]={0,1,0,-1};
int dy[4]={1,0,-1,0};
float sqr(float a) {
	return a*a;
}

float dist(const CPlayer &a, const CMonster &b) {
	return sqrt(sqr(a.PositionX-b.PositionX)+sqr(a.PositionY-b.PositionY));
} 
struct point {
	float x,y;
	point (float a=0, float b=0) {
		x=a;
		y=b;
	}
};

float dist(const point &a, const point &b) {
	return sqrt(sqr(a.x-b.x)+sqr(a.y-b.y));
} 
const point Zero;

struct cell {
	int monsters;
	float badness, badness2;
	float pathBadness;
	point prev;
};
cell cells[grid][grid];
bool b[grid][grid];
bool in(int i, int j) {
	return (i>=0 && i<grid && j>=0 && j<grid);
}
point getpath(int ki, int kj, int pi, int pj, int depth) {
	if (cells[ki][kj].prev.x==pi && cells[ki][kj].prev.y==pj || depth>=4*grid)
		return point(ki,kj); else
		return getpath(cells[ki][kj].prev.x,cells[ki][kj].prev.y,pi,pj,depth+1);
}
struct bfsTask
{
	int i, j;
	float badness;
	inline bfsTask(int _i, int _j, float _badness = 0.0f) : i(_i), j(_j), badness(_badness) { }
};

point chooseDirection(point p, point m, point mv, float sv)
{
	float t1, t2, l, r, mid, dx, dy, d, mx, my;
	point dp, sp;
//cerr << "~~" << mv.x << ' ' << mv.y << endl;
	if (mv.x == 0 && mv.y == 0) return point(m.x - p.x, m.y - p.y);
//cerr << "~" << endl;
	
	if (mv.x < 0) t1 = - m.x / mv.x;
	else if (mv.x > 0) t1 = (1 - m.x) / mv.x;
	else t1 = inf;

	if (mv.y < 0) t2 = - m.y / mv.y;
	else if (mv.y > 0) t2 = (1 - m.y) / mv.y;
	else t2 = inf;

	if (t1 < t2) r = t1; else r = t2;

	l = 0;
//cerr << r << endl;
	for (int i = 0; i < 100; ++i)
	{
		mid = (r + l) / 2;
		mx = m.x + mv.x * mid;
		my = m.y + mv.y * mid;

//cerr << mx << ' ' << my << endl;
		dx = mx - p.x;
		dy = my - p.y;
		d = sqrt(dx * dx + dy * dy);
		dp.x = dx * sv / d * mid;
		dp.y = dy * sv / d * mid;
		
		sp.x = mx - p.x;
		sp.y = my - p.y;
if (dist(dp, Zero) == dist(sp, Zero)) return dp;
		if (dist(dp, Zero) > dist(sp, Zero)) r = mid;
		else l = mid;
	}
	return dp;
}

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));
	float alpha = 0.0;
	float beta = 0.0;
	int ti=-1,tj;
	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		for (int i = 0; i < grid; i++)
			for (int j = 0; j < grid; j++)
			{
				cells[i][j].monsters = 0;
				cells[i][j].badness2 = 0.0;
			}

		for (int i = 0; i < world->MAXMONSTERS; i++)
		{
			if (world->monsters[i].ClassName >= 0)
			{
				int ci = world->monsters[i].PositionX * grid;
				int cj = world->monsters[i].PositionY * grid;
				ci = min(ci, grid - 1);
				cj = min(cj, grid - 1);
				cells[ci][cj].monsters++;
			}
		}

		for (int i = 0; i < grid; i++)
			for (int j = 0; j < grid; j++)
			{
				cells[i][j].badness2 = cells[i][j].monsters;
			}
		int nc;
		for (int i = 0; i < grid; i++)
			for (int j = 0; j < grid; j++)
			{
				nc=0;			
				for (int ti=i-1; ti<=i+1; ++ti)
					for (int tj=j-1; tj<=j+1; ++tj) if (in(ti,tj)) {
						++nc;
						cells[i][j].badness += cells[ti][tj].badness2;
					}
				cells[i][j].badness/=nc;
			}
		int pi, pj;
		pi = world->player.PositionX * grid;
		pj = world->player.PositionY * grid;
		pi = min(pi, grid - 1);
		pj = min(pj, grid - 1);
		int i,j,k,t;
		for (i=0; i<world->MAXMISSILES; ++i) if (world->missiles[i].ClassName>=0) {
			int ci = world->missiles[i].PositionX * grid;
			int cj = world->missiles[i].PositionY * grid;
			ci = min(ci, grid - 1);
			cj = min(cj, grid - 1);
			cells[ci][cj].badness+=1; //    
		}
		for (i=0; i<grid; ++i)
				for (j=0; j<grid; ++j)
						cells[i][j].pathBadness = inf;
		cells[pi][pj].pathBadness = cells[pi][pj].badness;
		for (i=0; i<grid; ++i)
				for (j=0; j<grid; ++j)
						b[i][j]=0;
		int ki,kj;
		while (1) {
			ki=-1;
			for (i=0; i<grid; ++i)
				for (j=0; j<grid; ++j)
					if (!b[i][j] && cells[i][j].pathBadness<inf && (ki==-1 || cells[i][j].pathBadness<cells[ki][kj].pathBadness)) {
							ki=i;
							kj=j;
					}
			b[ki][kj]=1;
			if (ki==-1) break;
			for (t=0; t<4; ++t) if (in(ki+dx[t],kj+dy[t]))
				if (cells[ki+dx[t]][kj+dy[t]].pathBadness>cells[ki][kj].pathBadness+cells[ki+dx[t]][kj+dy[t]].badness) {
					cells[ki+dx[t]][kj+dy[t]].pathBadness=cells[ki][kj].pathBadness+cells[ki+dx[t]][kj+dy[t]].badness;
					cells[ki+dx[t]][kj+dy[t]].prev=point(ki,kj);
				}
		}
		for (i=0; i<grid; ++i)
			for (j=0; j<grid; ++j)
				if (abs(i-pi)+abs(j-pj)>=grid/2+1)
					if (ti==-1 || (cells[i][j].pathBadness<cells[ti][tj].pathBadness
						|| fabs(cells[i][j].pathBadness-cells[ti][tj].pathBadness)<=eps && abs(i-pi)+abs(j-pj)>abs(pi-ti)+abs(pj-tj))) {
							ti=i;
							tj=j;
					}
//		cerr << "Here!\n";
		point pnext=getpath(ti,tj,pi,pj,1);
					//Run in circles (set direction)
		world->run(pnext.x-pi,pnext.y-pj);
		k=-1;
		//Choose the first monster that is still alive
		for ( int j = 0; j < world->MAXMONSTERS; j++)
			if (world->monsters[j].ClassName >= 0) {
				if (k==-1 || dist(world->player,world->monsters[j])<dist(world->player,world->monsters[k])) k=j;
				//Shoot the bastard! (Shoot in the direction of the monster)
//				world->shoot(world->monsters[j].PositionX - world->player.PositionX,
//						   world->monsters[j].PositionY - world->player.PositionY);
//				break;
			}
		if (k>=0) {
			//cerr << world->monsters[k].VelocityX << ' '<<world->monsters[k].VelocityY << endl;
			point fire=chooseDirection(point(world->player.PositionX,world->player.PositionY),
							   point(world->monsters[k].PositionX,world->monsters[k].PositionY),
							   point(world->monsters[k].VelocityX,world->monsters[k].VelocityY),
							   world->player.BulletVelocity);
//		cerr << fire.x << ' ' << fire.y << endl;
			world->shoot(fire.x,fire.y);
		}
	}

	return 0;
}
