#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <algorithm>
#include <vector>
#include <iostream>
#include <cctype>

using namespace std;

typedef double real;
const real PI = 3.1415926535897932384626433832795;
const int inf = 0x3f3f3f3f;
const int MAXMONSTERS = 256;
const real eps = 1e-5;
struct point{
	real x,y;
	point(){}
	point(real a, real b):x(a),y(b){}
	point perp()const{
		return point(y,-x);
	}
	real len2()const{
		return x*x+y*y;
	}
	real len()const{
		return sqrt(len2());
	}
	void norm(){
		real l = len();
		assert(l > eps);
		x /= l;
		y /= l;
	}
};

point operator+(const point& a, const point& b){
	return point(a.x+b.x,a.y+b.y);
}

point operator-(const point& a, const point& b){
	return point(a.x-b.x,a.y-b.y);
}

point operator*(const point& a, real k){
	return point(a.x*k,a.y*k);
}

ostream& operator<<(ostream& os, const point& p){
	return (os << p.x << ' ' <<p.y);
}

bool exists_shoooting_monster(World* world){
	for (int i = 0; i < world->MAXMONSTERS; i++){
		if (world->monsters[i].ClassName >= 0) {
			CMonster& monster = world->monsters[i];
			if (monster.BulletDamage){
				return true;
			}
		}
	}
	return false;
}

int bullets_shoot_in_monster[MAXMONSTERS];

static const int NUM_SEGMENTS_TO_CHECK = 36;
static const real SEGMENT_ANGLE = (2*PI)/NUM_SEGMENTS_TO_CHECK;
real segments[NUM_SEGMENTS_TO_CHECK];
static const int NUM_KOEFF = 3;
static const real KOEFF[NUM_KOEFF] = {1,1,0.85};//,0.5,0.25};

void calc_okrujenie(World* world){
	memset(segments,0,sizeof(segments));
	CPlayer& player = world->player;
	for (int i = 0; i < world->MAXMONSTERS; i++){
		if (world->monsters[i].ClassName >= 0) {
			CMonster& monster = world->monsters[i];
			real dy = monster.PositionY-player.PositionY;
			real dx = monster.PositionX-player.PositionX;
			if (point(dx,dy).len() > 0.33) continue;
			real angle = atan2(dy,dx);
			if (angle < 0) angle += 2*PI;
			int segm = int(angle/SEGMENT_ANGLE);
			segments[segm] += 1-point(dx,dy).len()*0.25;
		}
	}
	static const real GAP=0.10;
	static const int MANY_MONSTERS = 100;
	for (int i =0 ; i < NUM_SEGMENTS_TO_CHECK; i++) {
		real ang = (i+0.5)*SEGMENT_ANGLE;
		real dx = cos(ang);
		if (dx > 0){
			real dd = 1-world->player.PositionX;
			if (dd < GAP) segments[i] += MANY_MONSTERS;
		} else if (dx < 0){
			real dd = world->player.PositionX;
			if (dd < GAP) segments[i] += MANY_MONSTERS;
		}
		real dy = sin(ang);
		if (dy > 0){
			real dd = 1-world->player.PositionY;
			if (dd < GAP) segments[i] += MANY_MONSTERS;
		} else if (dy < 0){
			real dd = world->player.PositionY;
			if (dd < GAP) segments[i] += MANY_MONSTERS;
		}
	}
	for (int i = 0; i < NUM_SEGMENTS_TO_CHECK; i++){
		printf("%d - cos:%lf, sin:%lf, val: %lf\n",i,cos(SEGMENT_ANGLE*(0.5+i)),sin(SEGMENT_ANGLE*(0.5+i)),segments[i]);
	}
	puts("");
}

point get_dir_breakthrough(World* world, int& ATOM){
	calc_okrujenie(world);
	real summed[NUM_SEGMENTS_TO_CHECK];
	for (int i = 0; i < NUM_SEGMENTS_TO_CHECK; i++){
		summed[i] = segments[i];
		for (int j = 1; j <= NUM_KOEFF; j++){
			summed[i] += segments[(i+j)%NUM_SEGMENTS_TO_CHECK]*KOEFF[j-1];
			summed[i] += segments[(i+2*NUM_SEGMENTS_TO_CHECK-j)%NUM_SEGMENTS_TO_CHECK]*KOEFF[j-1];
		}
	}
	puts("summed");
	for (int i = 0; i < NUM_SEGMENTS_TO_CHECK; i++){
		printf("%d - cos:%lf, sin:%lf, val: %lf\n",i,cos(SEGMENT_ANGLE*(0.5+i)),sin(SEGMENT_ANGLE*(0.5+i)),summed[i]);
	}
	int best = 0;
	for (int i = 0; i < NUM_SEGMENTS_TO_CHECK; i++)
		if (summed[i] < summed[best])
			best = i;
	real ang = (best+0.5)*SEGMENT_ANGLE;
	printf("Best: %d Number: %lf\n",best,summed[best]);
	return point(cos(ang),sin(ang));
}

int get_need_bullets(World* world, CMonster& monster){
	return int(monster.MaxHitPoints/world->player.BulletDamage)+2;
}

int get_nearest(World* world){
	int best = -1;
	real mndist = inf;
	for (int i = 0; i < world->MAXMONSTERS;i++){
		CMonster& m = world->monsters[i];
		if (m.ClassName < 0) continue;
		int need_bullets = get_need_bullets(world,m);
		if (bullets_shoot_in_monster[i] >= need_bullets) continue;
		point dd = point(m.PositionX-world->player.PositionX,m.PositionY-world->player.PositionY);
		real dist = dd.len();
		if (dist < mndist || best < 0){
			best = i;
			mndist = dist;
		}
	}
	return best;
}

void do_shoot_at_monster(World* world, int best){
	if (best < 0) return;
	bullets_shoot_in_monster[best]++;
	CMonster& m = world->monsters[best];
	point to_shoot;
	point vel(world->player.VelocityX,world->player.VelocityY);
	point mpos(m.PositionX,m.PositionY);
	point pos(world->player.PositionX,world->player.PositionY);
	point mvel(m.VelocityX,m.VelocityY);
	point origpos = pos;
	real time_passed = 0;
	real bulletSpeed = world->player.BulletVelocity;
	static const real TIME_DELTA = 0.001;
	for (int t = 0; t < 10000; t++, time_passed += TIME_DELTA){
		point dd(mpos-origpos);
		if (dd.len()/bulletSpeed <= time_passed){
			break;
		}
		pos = pos+vel*TIME_DELTA;
		if (m.BulletDamage > 0){
			mpos = mpos + mvel*TIME_DELTA;
		} else {
			point tdd(pos-mpos);
			tdd.norm();
			tdd = tdd*m.MaxVelocity;
			mpos = mpos + tdd*TIME_DELTA;
		}
	}
	to_shoot = point(mpos-origpos);
	world->shoot(to_shoot.x,to_shoot.y);
}

void do_breakthrough(World* world, point dir){
	world->run(dir.x,dir.y);
	real angle = atan2(dir.y,dir.x);
	int best = -1;
	real mndist = inf;
	for (int i = 0; i < world->MAXMONSTERS;i++){
		CMonster& m = world->monsters[i];
		if (m.ClassName < 0) continue;
		int need_bullets = get_need_bullets(world,m);
		if (bullets_shoot_in_monster[i] >= need_bullets) continue;
		point dd = point(m.PositionX-world->player.PositionX,m.PositionY-world->player.PositionY);

		real ma = atan2(dd.y,dd.x);
		if (ma > angle+PI) ma -= 2*PI;
		if (ma < angle-PI) ma += 2*PI;

		static const real SECTOR = PI/6;
		if (abs(ma-angle) > SECTOR) continue;

		real dist = dd.len();
		if (best < 0 || dist < mndist){
			best = i;
			mndist = dist;
		}
	}
	if (best < 0) best = get_nearest(world);
	do_shoot_at_monster(world,best);
}

point get_monster_center(World* world){
	real x = 0;
	real y = 0;
	int cnt = 0;
	for (int i = 0; i < world->MAXMONSTERS; i++){
		CMonster& monster = world->monsters[i];
		if (monster.ClassName < 0) continue;
		x += monster.PositionX;
		y += monster.PositionY;
		cnt++;
	}
	return point(x/cnt,y/cnt);
}

void do_run_and_shoot(World* world, point dir){
	world->run(dir.x,dir.y);
	do_shoot_at_monster(world,get_nearest(world));
}

enum{
	UNKNOWN = 0,
	OKRUJ = 1,
	ROUND = 2,
};

void play_melee(World* world){
	int ATOM = 150;
	int startcount = 0;
	int cnt = 0;
	point dir_run(-inf,-inf);
	int state = UNKNOWN;
	for (;;cnt++){
		if (!(cnt%30)){
			memset(bullets_shoot_in_monster,0,sizeof(bullets_shoot_in_monster));
		}
		world->makeMove();
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);
		if (exists_shoooting_monster(world)) {
			printf("Shooting monster detected!!\n");
			ATOM = 100;
		} else {
			ATOM = 120;
		}

		if (cnt >= startcount+ATOM){
			state = UNKNOWN;
			printf("Reset!!!!!\n");
		}

		if (state == UNKNOWN){
			bool did = false;
			if (exists_shoooting_monster(world)) {
				point p(world->player.PositionX-0.5,world->player.PositionY-0.5);
				if (p.len() > 0.3){
					dir_run = point(-p.x,-p.y);
					did = true;
				}
			}
			if (!did){
				dir_run = get_dir_breakthrough(world,ATOM);
			}
			startcount = cnt;
			state = OKRUJ;
			cout << "running there: " << dir_run << endl;
		}

		do_breakthrough(world, dir_run);
	}
}

void play_shoot_and_kill(World* world){

}

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));

	play_melee(world);
	play_shoot_and_kill(world);
	return 0;
}
