#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>

#define sqr(a) ((a)*(a))
#define pi 3.1415926535897932384626433832795	
#define eps 1e-9
#define INF 200


double ArcTan(double x, double y, double bet)
{
	double ang = atan2(y, x);
	ang -= bet;
	if (ang < -eps) ang += 2 * pi;	
	return ang;
}

int bullets[100500];

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));
	float alpha = 0.0;
	float x = 0.46;


	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();
		//world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);
		int i;
		float res = 200;
		int ri = -1;
		double t;
		double bet;
		

		if (world -> player.Time < 165)
		{
			world -> run(1.0, 0);
			bet = 0.0;
		}
		else
		{
			t = (world -> player.Time - 165.0) / 175.0;
			//world -> run(exp(x * t) * (x * cos(t) - sin(t)), exp(x * t) * (x * sin(t) + cos(t)));
			world -> run(-sin(t), cos(t));
			bet = atan2(cos(t), -sin(t));
		}

		if (true)
		{
			res = INF;
			double hit = INF;
			for (i = 0; i < world -> MAXMONSTERS; i++)
			{	
				if (world -> monsters[i].ClassName < 0) continue;
				//printf("angl: %.2lf dist:%.2lf r:%.8lf\n", (ArcTan(world -> monsters[i].PositionX - world -> player.PositionX, world -> monsters[i].PositionY - world -> player.PositionY, bet)) * 180 / pi,
				//sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY),
				//sqr(world -> monsters[i].Radius + world -> player.Radius));
			    
			    if ((ArcTan(world -> monsters[i].PositionX - world -> player.PositionX, world -> monsters[i].PositionY - world -> player.PositionY, bet) < pi / 8.0 - eps
			    || ArcTan(world -> monsters[i].PositionX - world -> player.PositionX, world -> monsters[i].PositionY - world -> player.PositionY, bet) > 2 * pi - pi / 8.0 + eps)
			    && sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY) 
			    < 100 * sqr(world -> monsters[i].Radius + world -> player.Radius) - eps
			    && (sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY) < res - eps
			    || fabs( sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY) - res) < eps
			    && world -> monsters[i].HitPoints < hit - eps))
			    {
			    	res = sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY);
			    	hit = world -> monsters[i].HitPoints;
			    	world -> shoot(world -> monsters[i].PositionX - world -> player.PositionX, world -> monsters[i].PositionY - world -> player.PositionY);
			    }
			}
			if (fabs(res - INF) > eps)
			{
				continue;
			}
			
			res = 200;
			for (i = 0; i < world -> MAXMONSTERS; i++)
			{
				//if (sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY) < res)
				if (world -> monsters[i].ClassName < 0) continue;
				if (ri == -1 || world -> monsters[i].HitPoints < res - 1e-6 || fabs(world -> monsters[i].HitPoints - res) < 1e-6
				&& sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY)
				< sqr(world -> monsters[ri].PositionX - world -> player.PositionX) + sqr(world -> monsters[ri].PositionY - world -> player.PositionY) )
				{
					res = world -> monsters[i].HitPoints;
					ri = i;
				//	res = sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY);
				}
			}
			
			
			if (ri != -1)
			{
			//	world -> run(world -> player.PositionX - world -> monsters[ri].PositionX, world -> player.PositionY - world -> monsters[ri].PositionY);
				world -> shoot(-world -> player.PositionX + world -> monsters[ri].PositionX, -world -> player.PositionY + world -> monsters[ri].PositionY);
			}
			else
			{
				world -> run(0,0);
			}
		}

		if (world -> player.Time >= 1800 && world -> player.Time < 3200)
		{
			for (i = 0; i < world -> MAXMONSTERS; i++)
			{
				if (world -> monsters[i].ClassName < 0)
				{
					bullets[i] = 0;
					continue;
				}
				if (bullets[i] > 5) continue;
				double vx = world -> monsters[i].VelocityX, vy = world -> monsters[i].VelocityY;
				double x0 = world -> monsters[i].PositionX, y0 = world -> monsters[i].PositionY;
				double b = world -> player.BulletVelocity;
				world -> shoot((world -> monsters[i].PositionX + world -> monsters[i].VelocityX * 20) - world -> player.PositionX, (world -> monsters[i].PositionY + world -> monsters[i].VelocityY * 20) - world -> player.PositionY);
			//	world -> shoot((x0 * b - world -> player.PositionX * vx) / (b - vx) - world -> player.PositionX, (y0 * b - world -> player.PositionY * vy) / (b - vy) - world -> player.Positiony);
				bullets[i]++;
				break;
			}
		}
	/*	else if (world -> player.Time < 1800)
		{
			for (i = 0; i < world -> MAXMONSTERS; i++)
			{
				//
				if (world -> monsters[i].ClassName < 0) continue;
			//	if (sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY) < res)
				if (ArcTan(world -> monsters[i].PositionX - world -> player.PositionX, world -> monsters[i].PositionY - world -> player.PositionY, bet) < res - eps)
				{
					res = ArcTan(world -> monsters[i].PositionX - world -> player.PositionX, world -> monsters[i].PositionY - world -> player.PositionY, bet);
					ri = i;
				//	res = sqr(world -> monsters[i].PositionX - world -> player.PositionX) + sqr(world -> monsters[i].PositionY - world -> player.PositionY);
				}
			}
			
			if (ri != -1)
			{
			//	world -> run(world -> player.PositionX - world -> monsters[ri].PositionX, world -> player.PositionY - world -> monsters[ri].PositionY);
				world -> shoot(-world -> player.PositionX + world -> monsters[ri].PositionX, -world -> player.PositionY + world -> monsters[ri].PositionY);
			}
			else
			{
				world -> run(0,0);
			}
		}
		*/

			double Bx, By;
			Bx = world -> player.PositionX;
			By = world -> player.PositionY;
			double R = world -> player.Radius;
			double A, B, C;
			double vx, vy, x0, y0;
			double r, t1;
			double rx = 0.0, ry = 0.0;
			for (i = 0; i < world -> MAXMISSILES; i++)
				if (world -> missiles[i].ClassName > 200)
				{
					A = -world -> missiles[i].VelocityY;
					B = world -> missiles[i].VelocityX;
					vx = world -> missiles[i].PositionX;
					vy = world -> missiles[i].PositionY;
					C = -(vx * A + vy * B);
					r = world -> missiles[i].Radius;
					t1 = -(A * Bx + B * By + C) / (A * A + B * B);
					if (sqrt(A * A + B * B) * fabs(t1) < 5*(r + R) + eps)
					{
						x0 = Bx + t1 * A;
						y0 = By + t1 * B;
						if (fabs(B) < eps)
						{
							t1 = (y0 - vy) / (-A);
						}
						else
						{
							t1 = (x0 - vx) / B;
						}
						if (t1 > -eps)
						{
							if ((A * Bx + B * By + C) < -eps)
							{
								rx -= A;
								ry -= B;
							}	
							else
							{
								rx += A;
								ry += B;
							}
						}
					}

				}
		   	if (sqr(rx) + sqr(ry) > eps)
		   		world -> run(rx, ry);
			
		
		

		/*
		alpha += 0.02;
		//Run in circles (set direction)
		world->run(cos(alpha), sin(alpha));

		//Choose the first monster that is still alive
		for ( int j = 0; j < world->MAXMONSTERS; j++)
			if (world->monsters[j].ClassName >= 0) {
				//Shoot the bastard! (Shoot in the direction of the monster)
				world->shoot(world->monsters[j].PositionX - world->player.PositionX,
						   world->monsters[j].PositionY - world->player.PositionY);
				break;
			}
		*/
	}

	return 0;
}
