#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <utility>

const float PI = 3.1415;
const int EDGE = 5;

using namespace std;

int tx = 1, ty = 1;
float dx = 0, dy = 0;
int cx = 0, cy = 0;

int near(World *world) {
	int m = -1;
	float d = 0;

	for(int i = 0; i < world->MAXMONSTERS; ++i)
		if(world->monsters[i].ClassName >= 0) {
			float dx = world->player.PositionX - world->monsters[i].PositionX;
			float dy = world->player.PositionY - world->monsters[i].PositionY;
			float dis = sqrt(dx * dx + dy * dy);

			if(m == -1 || dis < d) {
				m = i;
				d = dis;
			}
		}

	return m;
}

pair<float, float> summ_all(World *world) {
	pair<float, float> res = make_pair(0, 0);

	for(int i = 0; i < world->MAXMONSTERS; ++i)
		if(world->monsters[i].ClassName >= 0) {
			res.first += world->monsters[i].VelocityX - world->player.PositionX;
			res.second += world->monsters[i].VelocityY - world->player.PositionY;
		}

	return res;
}

int main(int argc, char** argv) {
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));
	float alpha = 0.0;

	/*
	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

		alpha += 0.02;
		//Run in circles (set direction)
		world->run(cos(alpha), sin(alpha));

		//Choose the first monster that is still alive
		for ( int j = 0; j < world->MAXMONSTERS; j++)
			if (world->monsters[j].ClassName >= 0) {
				//Shoot the bastard! (Shoot in the direction of the monster)
				world->shoot(world->monsters[j].PositionX - world->player.PositionX,
						   world->monsters[j].PositionY - world->player.PositionY);
				break;
			}
	}
	*/

	int r = 0;

	while(true) {
		world->makeMove();

		if(dx == world->player.PositionX)
			cx++;
		else
			cx = 0;

		if(dy == world->player.PositionY)
			cy++;
		else
			cy = 0;

		if(cx > EDGE) {
			tx = -tx;
			cx = 0;
		}
		
		if(cy > EDGE) {
			ty = -ty;
			cy = 0;
		}
		
		pair<float, float> d = summ_all(world);
		world->run(d.first * tx, d.second * ty);
		
		int m = near(world);
		if(m != -1)
			world->shoot(world->monsters[m].PositionX - world->player.PositionX,
						 world->monsters[m].PositionY - world->player.PositionY);
		
		dx = world->player.PositionX;
		dy = world->player.PositionY;
	}

	return 0;
}
