#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <iostream>
#include <iomanip>
#include <sstream>

#include <string>
#include <algorithm>
#include <set>
#include <map>
#include <vector>
#include <utility>

template<typename tt>
tt sqr(const tt &x) {
	return x * x;
}

#define DB(x) std::cout << #x << " = " << x << '\n'
#define SIZE(x) (sizeof(x) / sizeof((x)[0]))
#define MP std::make_pair

const float pi = acos(-1.F);

typedef std::pair<float, float> pff;

std::vector<pff> waypoints;

float Cos(float x1, float y1, float x2, float y2) {
	float res = x1 * x2 + y1 * y2;
	res /= sqrt(sqr(x1) + sqr(y1));
	res /= sqrt(sqr(x2) + sqr(y2));

	return res;
}

//#define DEBUG

int main(int argc, char** argv) {
	srand(777);

#ifndef DEBUG
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	World *world = new World(argv[1], atoi(argv[2]));

#else
	World *world = new World("127.0.0.1", 6666);
#endif

	const int pCnt = 18;
	for (int i = 0; i < pCnt; ++i) {
		const float r = 0.4F;
		float alpha = 2 * pi / pCnt * i;

		waypoints.push_back(MP(r * cos(alpha) + 0.5F, r * sin(alpha) + 0.5F));
	}

	int curW = -1;

	while (true) {
		world->makeMove();

		//Print some information
		/*printf("frame: %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);*/

		float minD = 100;
		int minIdx = -1;

		for (int j = 0; j < world->MAXMONSTERS; j++) {
			if (world->monsters[j].ClassName >= 0) {
				float curD = sqrt(sqr(world->monsters[j].PositionX - world->player.PositionX) + sqr(world->monsters[j].PositionY - world->player.PositionY));
				curD /= world->monsters[j].MaxVelocity;
				
				if (minIdx == -1 || curD < minD) {
					minD = curD;
					minIdx = j;
				}
			}
		}

		DB(minD);

		if (/*minD < 30.F / 60*/ false) {
			DB("alarm!");
			world->shoot(world->monsters[minIdx].PositionX - world->player.PositionX,
				world->monsters[minIdx].PositionY - world->player.PositionY);
			world->run(-(world->monsters[minIdx].PositionX - world->player.PositionX),
				-(world->monsters[minIdx].PositionY - world->player.PositionY));

		} else {
			//move
			float toNextWP = sqrt(sqr(waypoints[(curW + 1) % waypoints.size()].first - world->player.PositionX) + 
				sqr(waypoints[(curW + 1) % waypoints.size()].second - world->player.PositionY));

			float runToX = waypoints[(curW + 1) % waypoints.size()].first - world->player.PositionX;
			float runToY = waypoints[(curW + 1) % waypoints.size()].second - world->player.PositionY;

			world->run(runToX, runToY);

			if (toNextWP < 0.02) {
				curW = (curW + 1) % waypoints.size();
			}
			
			//shoot
			float cx = 0, cy = 0;
			int mAlive = 0;
			float normCoef = 0;

			for (int j = 0; j < world->MAXMONSTERS; j++) {
				if (world->monsters[j].ClassName >= 0) {
					float d = sqrt(sqr(world->monsters[j].PositionX - world->player.PositionX) + sqr(world->monsters[j].PositionY - world->player.PositionY)) + 0.01F;
					float angleToMonster = Cos(runToX, runToY, world->monsters[j].PositionX - world->player.PositionX, world->monsters[j].PositionY - world->player.PositionY);

					if (angleToMonster < 0.85 /*&& d > 0.25*/)
						continue;
					
					cx += world->monsters[j].PositionX * (1 / d);
					cy += world->monsters[j].PositionY * (1 / d);
					++mAlive;
					normCoef += (1 / d);
				}
			}

			cx /= normCoef;
			cy /= normCoef;

			int optMonster = -1;
			float optd = 2;

			for (int j = 0; j < world->MAXMONSTERS; j++) {
				if (world->monsters[j].ClassName >= 0) {
					float curd = sqrt(sqr(world->monsters[j].PositionX - cx) + sqr(world->monsters[j].PositionY - cy));
					
					if (curd < optd) {
						optd = curd;
						optMonster = j;
					}
				}
			}

			//check some classes
			bool isStayShooter = false;
			
			for (int j = 0; j < world->MAXMONSTERS; j++) {
				if (world->monsters[j].ClassName >= 0) {
					float vel = sqrt(sqr(world->monsters[j].VelocityX) + sqr(world->monsters[j].VelocityY));
					
					if (world->monsters[j].BulletDamage > 0 && vel < 0.01) {
						DB(world->monsters[j].BulletDamage);
						isStayShooter = true;
					}
				}
			}

			if (optMonster == -1) {
				if (isStayShooter) {
					optMonster = -1;
					float optd = 2;

					for (int j = 0; j < world->MAXMONSTERS; j++) {
						if (world->monsters[j].ClassName >= 0) {
							float curd = sqrt(sqr(world->monsters[j].PositionX - world->player.PositionX) + sqr(world->monsters[j].PositionY - world->player.PositionY));
							float vel = sqrt(sqr(world->monsters[j].VelocityX) + sqr(world->monsters[j].VelocityY));
							
							if (world->monsters[j].BulletDamage > 0 && vel < 0.01 && curd < optd) {
								optd = curd;
								optMonster = j;
							}
						}
					}

				} else {
					cx = 0;
					cy = 0;
					float normCoef = 0;

					for (int j = 0; j < world->MAXMONSTERS; j++) {
						if (world->monsters[j].ClassName >= 0) {
							float d = sqrt(sqr(world->monsters[j].PositionX - world->player.PositionX) + sqr(world->monsters[j].PositionY - world->player.PositionY)) + 0.01F;
							float angleToMonster = Cos(runToX, runToY, world->monsters[j].PositionX - world->player.PositionX, world->monsters[j].PositionY - world->player.PositionY);

							cx += world->monsters[j].PositionX /** (1 / d)*/ * (1 / world->monsters[j].HitPoints);
							cy += world->monsters[j].PositionY /** (1 / d)*/ * (1 / world->monsters[j].HitPoints);
							normCoef += /*(1 / d) **/ (1 / world->monsters[j].HitPoints);
						}
					}

					cx /= normCoef;
					cy /= normCoef;

					optMonster = -1;
					float optd = 2;

					for (int j = 0; j < world->MAXMONSTERS; j++) {
						if (world->monsters[j].ClassName >= 0) {
							float curd = sqrt(sqr(world->monsters[j].PositionX - cx) + sqr(world->monsters[j].PositionY - cy));
							
							if (curd < optd) {
								optd = curd;
								optMonster = j;
							}
						}
					}
				}
			}

			if (optMonster == -1) {
				DB("no monster");
				world->run(0.5F - world->player.PositionX, 0.5F - world->player.PositionY);

				//shoot
				float shY = float(rand()) / RAND_MAX * 0.3F + 0.3F;
				world->shoot(0 - world->player.PositionX, shY - world->player.PositionY);

				continue;
			}

			world->shoot(world->monsters[optMonster].PositionX - world->player.PositionX,
				world->monsters[optMonster].PositionY - world->player.PositionY);
		}
	}

	return 0;
}
