#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <algorithm>
#define INF 1000*1000*1000

#define mp make_pair
#define sc second
#define fs first
#define pb push_back

using namespace std;

const int forward_move = 160;
const double PI = 3.1415926535897932384626433832795;
const double pi = 3.1415926535897932384626433832795;
const double inf = 1e8;

const double eps = 1e-7;
const int size = 256;
const double rad = 2.2;
const double min_dist = 1e-1;

double alpha = 0;
int steps_done = 0, world_time = 0;
double dam, h, lovex, lovey;
double BulDam[size], BulHP[size], BulVel[size], BulRad[size], monster_damage[size];
pair <double, int> monanddist[size];

struct point
{
    double x;
    double y;
};

double DistPointToPoint(point p1, point p2)
{
    return sqrt((p1.x - p2.x) * (p1.x - p2.x) + (p1.y - p2.y) * (p1.y - p2.y));
}

double DistToPoint(double x1, double y1, double x2, double y2)
{
    double xd = (x1 - x2) * (x1 - x2);
    double yd = (y1 - y2) * (y1 - y2);
    return sqrt(xd + yd);
}

double DistToMonster(int i, World* world)
{
    return DistToPoint(world -> player.PositionX, world -> player.PositionY, world -> monsters[i].PositionX, world -> monsters[i].PositionY);
}
void ShootStraightToMonster(int i, World* world)
{
    world->shoot(world->monsters[i].PositionX - world->player.PositionX, world->monsters[i].PositionY - world->player.PositionY);
}

void CircleMoving(World* world)
{
    if (world -> player.Time < forward_move)
    {
        world->run(1, 0);
        return;
    }
    //world_time != world -> player.Time &&
    if ((DistToPoint(world -> player.PositionX, world -> player.PositionY, cos(alpha) / rad + 0.5, sin(alpha) / rad + 0.5) < min_dist))
        alpha += 0.006;
    world_time = world -> player.Time;
    //Run in circles (set direction)
	world->run(cos(alpha) / rad + 0.5 - world -> player.PositionX, sin(alpha) / rad + 0.5 - world -> player.PositionY);
	lovex = cos(alpha) / rad + 0.5 - world -> player.PositionX;
	lovey = sin(alpha) / rad + 0.5 - world -> player.PositionY;
}


int ChooseNearest(World* world)
{
    int k = -1;
    for ( int j = 0; j < world->MAXMONSTERS; j++)
			if ((k==-1) || (DistToMonster(j, world) < DistToMonster(k, world))) k=j;
    return k;
}

double BulletDam(World* world)
{
    for (int i = 0; i < world->MAXMONSTERS; i++)
        if (world -> monsters[i].ClassName > 0)
            return world -> monsters[i].BulletDamage;
    return 0;
}

double BulletHel(World* world)
{
    for (int i = 0; i < world -> MAXMONSTERS; i++)
        if (world -> monsters[i].ClassName > 0)
            return world -> monsters[i].BulletHitPoints;
    return 0;
}
bool pointInBoard(double x, double y)
{
    return (x>=0)&&(x<=1)&&(y>=0)&&(y<=1);
}

double TimeOfMinimalDistance(double x0, double y0, double x1, double y1, double vx0, double vy0, double vx1, double vy1)
{
    return ((x1 - x0) * (vx1 - vx0) + (y1 - y0) * (vy1 - vy0)) / (((vx1 - vx0) * (vx1 - vx0)) + (vy1 - vy0) * (vy1 - vy0));
}

point MonsterWillBe(int mn, double mtm, World* world)
{
    point p;
    p.x = world -> monsters[mn].PositionX + world -> monsters[mn].VelocityX * mtm;
    p.y = world -> monsters[mn].PositionY + world -> monsters[mn].VelocityY * mtm;
    return p;
}

point BulletWillBe(int bl, double mtm, World* world)
{
    point p;
    p.x = world -> missiles[bl].PositionX + world -> missiles[bl].VelocityX * mtm;
    p.y = world -> missiles[bl].PositionY + world -> missiles[bl].VelocityY * mtm;
    return p;
}

bool BulletMeetsMonster(int mn, int bl, World* world)
{
    double x0 = world -> monsters[mn].PositionX, y0 = world -> monsters[mn].PositionY;
    double x1 = world -> missiles[bl].PositionX, y1 = world -> missiles[bl].PositionY;
    double vx0 = world -> monsters[mn].VelocityX, vy0 = world -> monsters[mn].VelocityY;
    double vx1 = world -> missiles[bl].VelocityX, vy1 = world -> monsters[mn].VelocityY;
    double mtm = max(0.0, TimeOfMinimalDistance(x0, y0, x1, y1,vx0, vy0, vx1, vy1));
    point p1 = MonsterWillBe(mn, mtm, world);
    point p2 = BulletWillBe(bl, mtm, world);
    return pointInBoard(p2.x, p2.y) && (DistPointToPoint(p1, p2) <= world -> monsters[mn].Radius + BulRad[world -> missiles[bl].ClassName - 100]);
}

bool ShootToMovingMonster(int i, World* world)
{
    double x0=world->monsters[i].PositionX, y0=world->monsters[i].PositionY,
           vx0=world->monsters[i].VelocityX, vy0=world->monsters[i].VelocityY,
           x=world->player.PositionX, y=world->player.PositionY;
    double v=world->player.BulletVelocity;
    double a=v*v-vx0*vx0-vy0*vy0, b=2*(vx0*(x-x0)+vy0*(y-y0)), c=-((x-x0)*(x-x0)+(y-y0)*(y-y0));
    double d=b*b-4*a*c;
    if(d>-eps)
    {
        double t1=(-b+sqrt(d+eps))/(2*a), t2=(-b-sqrt(d+eps))/(2*a);
        if((t1<-eps)||(t2>-eps)&&(t2<t1)) swap(t1, t2);
        if((t1>-eps)&&(pointInBoard(x0+vx0*t1, y0+vy0*t1)))
        {
            double newx=x0+vx0*t1, newy=y0+vy0*t1;
            world->shoot(newx-x, newy-y);
            printf("Ok!!!\n");
            return true;
        }
        swap(t1, t2);
        if((t1>-eps)&&(pointInBoard(x0+vx0*t1, y0+vy0*t1)))
        {
            double newx=x0+vx0*t1, newy=y0+vy0*t1;
            world->shoot(newx-x, newy-y);
            printf("Ok!!!\n");
            return true;
        }
    }
    ShootStraightToMonster(i, world);
    return false;
}


double CalculateOneMonsterAndBullets(int mn, World* world)
{
    double dm = 0;
    for (int i = 0; i < world -> MAXMISSILES; i++)
        if (world -> missiles[i].ClassName > 0 && BulletMeetsMonster(mn, i, world))
            dm += BulDam[world -> missiles[i].ClassName - 100];
    return dm;

}

void CalculateBullets(World* world)
{
    int i, j;
    for (i = 0; i < world -> MAXMONSTERS; i++)
        if (world -> monsters[i].ClassName > 0)
            monster_damage[i] = CalculateOneMonsterAndBullets(i, world);
        else
            monster_damage[i] = 0;
}
const int nparts=100;
point p[nparts];
int cnt[nparts];
point mPoint(double x, double y)
{
    point res;
    res.x=x;
    res.y=y;
    return res;
}
double Vmul(point p1, point p2)
{
    return p1.x*p2.y-p1.y*p2.x;
}
point v(point p1, point p2)
{
    point res;
    res.x=p2.x-p1.x;
    res.y=p2.y-p1.y;
    return res;
}


double TimeToReach(int mn, World* world)
{
    double x = world -> player.PositionX;
    double y = world -> player.PositionY;
    double xh = world -> monsters[mn].PositionX;
    double yh = world -> monsters[mn].PositionY;
    double vx = lovex;
    double vy = lovey;
    double vxi = world -> player.VelocityX;
    double vyi = world -> player.VelocityY;
    double sp = sqrt(vxi * vxi + vyi * vyi);
    double tlen = sqrt(vx * vx + vy * vy);
    vx = vx * (sp / tlen);
    vy = vy * (sp / tlen);
    double u = world -> monsters[mn].MaxVelocity;
    double a = (vx  * vx + vy * vy), b = 2 * vx * (x - xh) + 2 * vy * (y - yh) - u, c = (x - xh) * (x - xh) + (y - yh) * (y - yh);
    double disk = b * b - 4 * a * c;
    if (disk < 0)
        return inf;
    double x1 = (- b + sqrt(disk)) / (2 * a), x2 = (-b - sqrt(disk)) / (2 * a);
    double ans = inf;
    if (x1 > 0 && x1 < ans) ans = x1;
    if (x2 > 0 && x2 < ans) ans = x2;
    return ans;
}
double cb(double x)
{return x*x*x;}

void DensityMoving(World* world)
{
    double alpha=2*PI/nparts;
    double x0=world->player.PositionX, y0=world->player.PositionY;
    int i=0, k=0;
    for(; i<nparts; ++i)
    {
        if(i*alpha>pi/2-eps) break;
        double ix=y0+(1-x0)*tan(i*alpha);
        if(ix>1+eps) break;
        p[k++]=mPoint(1, ix);
    }
    for(; i<nparts; ++i)
    {
        if(i*alpha>pi-eps) break;
        double ix=x0+(1-y0)*tan((nparts/4 - i)*alpha);
        if(ix<-eps) break;
        p[k++]=mPoint(ix, 1);
    }
    for(; i<nparts; ++i)
    {
        if(i*alpha>3*pi/2 - eps) break;
        double ix=y0-x0*tan(i*alpha);
        if(ix<-eps) break;
        p[k++]=mPoint(0, ix);
    }
    for(; i<nparts; ++i)
    {
        double ix=x0-y0*tan((3*nparts/4 - i)*alpha);
        if(ix>1+eps) break;
        p[k++]=mPoint(ix, 0);
    }
    for(; i<nparts; ++i)
    {
        double ix=y0+(1-x0)*tan(i*alpha);
        p[k++]=mPoint(1, ix);
    }
    int angle;
    for(i=0; i<world->MAXMONSTERS; ++i)
        if(world->monsters[i].ClassName>0)
        {
            double mycos=(world->monsters[i].PositionX-x0)/DistToMonster(i, world);
            double mysin=(world->monsters[i].PositionY-y0)/DistToMonster(i, world);
            double anglecos=acos(mycos);
            double anglesin=asin(mysin);
            if(mysin>0) angle=(int) floor(anglecos/alpha+eps); else angle=(int) floor((pi+anglecos)/alpha+eps);
            cnt[angle]++;
        }
    point tp=mPoint(x0, y0);
    int j=0; double best=1e100;
    for(i=0; i<nparts-1; ++i)
    {
        double vm=abs(Vmul(v(tp, p[i]), v(tp, p[i+1]))/2);
        if((cnt[i]+1)/cb(vm+eps)<best) best=(cnt[i]+1)/cb(vm+eps), j=i;
    }
    lovex=cos(alpha*j); lovey=sin(alpha*j);
    world->run(lovex, lovey);
}

void GetTargetAndFire(World* world)
{
    int i;
    //CalculateBullets(world);
    //int k = ChooseNearest(world);
    int len = 0;
    for (i = 0; i < world -> MAXMONSTERS; i++)
        if (world -> monsters[i].ClassName > 0)
            monanddist[len++] = mp(DistToMonster(i, world), i);
            //monanddist[len++] = mp(TimeToReach(i, world), i);
    sort(monanddist, monanddist + len);
    if (steps_done == 200)
    {
        for (i = 0; i < len; i++)
            printf("%.5lf %d\n", monanddist[i].fs, monanddist[i].sc);
    }

        for (i = 0; i < len; i++)
            if ( (world -> monsters[monanddist[i].sc].HitPoints - monster_damage[monanddist[i].sc]) > eps && ShootToMovingMonster(monanddist[i].sc, world))


                return;

    for (i = 0; i < len; i++)
        if (ShootToMovingMonster(monanddist[i].sc, world))
            return;
}

void Moving(World* world)
{
    //DensityMoving(world);
    CircleMoving(world);
}


void CalculateClassesHits(World* world)
{
    int i, j;
    for (i = 0; i < world -> MAXMONSTERS; i++)
        if (world -> monsters[i].ClassName > 0)
        {
            BulDam[world -> monsters[i].ClassName] = world -> monsters[i].BulletDamage;
            BulHP[world -> monsters[i].ClassName] = world -> monsters[i].BulletHitPoints;
            BulVel[world -> monsters[i].ClassName] = world -> monsters[i].BulletVelocity;
            BulRad[world -> monsters[i].ClassName] = world -> monsters[i].BulletRadius;
        }
}


int main(int argc, char** argv) {
	//Check for enough parameters
	freopen("report.txt", "w", stdout);
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	World *world = new World(argv[1], atoi(argv[2]));

	//Infinite loop
	while (true) {
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		printf("frame : %d, HP: %.2lf x: %.2lf y:%.2lf\n",
				world->player.Time, world->player.HitPoints,
				world->player.PositionX, world->player.PositionY);

        CalculateClassesHits(world);
        Moving(world);
        GetTargetAndFire(world);

		if (world -> player.Time == 2000)
		{
            dam = BulletDam(world);
            h = BulletHel(world);
            printf("MAXMONSTERS = %d\n", world -> MAXMONSTERS);
            printf("BulletDamage = %.5lf\n", dam);
            printf("BulletHealth = %.5lf\n", h);
		}

		//Choose the first monster that is still alive
		//ShootToMovingMonster(world, k);
		//if(k!=-1) ShootStraightToMonster(k, world);
        steps_done++;
	}



	return 0;
}

