#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <iostream>
#include <vector>
#include <algorithm>
using namespace std;
#define forn(i,n) for(int i=0;i<n;++i)
#define all(v) v.begin(),v.end()
#define SQ(a) ((a)*(a))
#define PI 3.1415926535897932384626433832795

const int ESCAPE = 1;
const int CLOUD = 2;
const int PANIC = 3;
const int SURVIVE = 4;
const int SNIPER = 5;
const int ACROSS = 6;
const int WAIT_INSIDE = 7;
const int UPDOWN = 8;

World *world;
int step;
int mode = 3;
float escape_x,escape_y;
float cloud_x,cloud_y;
float goal_x,goal_y;

int shot[1000000];
const int shot_retry_time=30;

bool can_shoot_at(int m){
	return shot[m]==0;
}
void mark_shot(int m){
	shot[m]=shot_retry_time;
}
void upd_shots(){
	forn(i,world->MAXMONSTERS)
		if(world->monsters[i].ClassName==-1)
			shot[i]=0;
		else
			shot[i]=max(0,shot[i]-1);
}

void my_shoot(float x,float y,float scatter){
	float d=sqrt(x*x+y*y);
	if(d==0)
		d=1;
	x/=d;
	y/=d;
	x+=rand()*1.0/(RAND_MAX-1)*scatter;
	y+=rand()*1.0/(RAND_MAX-1)*scatter;
	world->shoot(x,y);
}

bool in_angle(int i,float a1,float range){
	float t=atan2(world->monsters[i].PositionY-world->player.PositionY,world->monsters[i].PositionX-world->player.PositionX);
	t-=a1;
	while(t<0)
		t+=2*PI;
	while(t>=2*PI)
		t-=2*PI;
	return t<=range;
}

int monsters(){
	int c=0;
	forn(i,world->MAXMONSTERS){
		if(world->monsters[i].ClassName==-1)
			continue;
		++c;
	}
	return c;
}
bool wave2(){
	forn(i,world->MAXMONSTERS){
		if(world->monsters[i].ClassName==-1)
			continue;
		if(world->monsters[i].BulletVelocity>0)
			return true;
	}
	return false;
}

const float shootdist=0.1;
int choose_shoot(float dx,float dy,float range){
	float a1=atan2(dy,dx)-range/2;

	int best=-1;
	pair<int,float> bestv(-1000000000,0);

	float x=world->player.PositionX;
	float y=world->player.PositionY;
	
	forn(i,world->MAXMONSTERS){
		if(world->monsters[i].ClassName==-1)
			continue;
		if(!in_angle(i,a1,range))
			continue;
		float d=sqrt(SQ(world->monsters[i].PositionX-x)+SQ(world->monsters[i].PositionY-y));
		pair<int,float> p;
		p.first=-shot[i]*2+((d<shootdist)?1:0);
		p.second=-world->monsters[i].HitPoints;
		if(p>bestv){
			bestv=p;
			best=i;
		}
	}
	return best;
}

void sniper_shot(int t){
	mark_shot(t);
	float mx=world->monsters[t].PositionX;
	float my=world->monsters[t].PositionY;
	mx-=world->player.PositionX;
	my-=world->player.PositionY;
	float vx=world->monsters[t].VelocityX;
	float vy=world->monsters[t].VelocityY;
	float V=world->player.BulletVelocity;
	float a=SQ(vx)+SQ(vy)-SQ(V);
	float b=mx*vx+my*vy;
	float c=SQ(mx)+SQ(my);
	float d=b*b-a*c;
	if(d<0)
		my_shoot(0,0,10);
	else{
		float tt;
		if(a>0)
			tt=(sqrt(d)-b)/a;
		else
			tt=(-sqrt(d)-b)/a;
		float dx=vx+mx/tt;
		float dy=vy+my/tt;
		world->shoot(dx,dy);
	}
}
void sniper_shot(){
	sniper_shot(choose_shoot(0,1,3*PI));
}

void aim_shoot(float dx,float dy,float range){
	int best=choose_shoot(dx,dy,range);
	sniper_shot(best);
	//mark_shot(best);
	//world->shoot(world->monsters[best].PositionX-world->player.PositionX,world->monsters[best].PositionY-world->player.PositionY);
}

const int out_circle = 40;
void get_circle(float &sx,float &sy,float &rad){
	sx=0,sy=0;
	int c=0;
	forn(i,world->MAXMONSTERS){
		if(world->monsters[i].ClassName==-1)
			continue;
		sx+=world->monsters[i].PositionX;
		sy+=world->monsters[i].PositionY;
		++c;
	}
	sx/=c;
	sy/=c;

	vector<float> dists;
	forn(i,world->MAXMONSTERS){
		if(world->monsters[i].ClassName==-1)
			continue;
		dists.push_back(SQ(world->monsters[i].PositionX-sx)+SQ(world->monsters[i].PositionY-sy));
	}
	sort(all(dists));
	if(!dists.size()){
		rad=0;
		return;
	}
	rad=sqrt(dists[max(0,(int)(dists.size()-out_circle-1))]);
}

void debug_out(){
	//cout<<world->player.RunX<<' '<<world->player.RunY<<' '<<world->player.ShootX<<' '<<world->player.ShootY<<' '<<endl;
	cout<<mode<<endl;
}

const float ang_range=PI/6;
void shoot_forward(float sc){
	//my_shoot(world->player.RunX,world->player.RunY,sc);
	aim_shoot(world->player.RunX,world->player.RunY,ang_range);
}
void shoot_cloud(float sc){
	float sx,sy,r;
	get_circle(sx,sy,r);
	//my_shoot(sx-world->player.PositionX,sy-world->player.PositionY,sc);
	aim_shoot(sx-world->player.PositionX,sy-world->player.PositionY,ang_range);
	//aim_shoot(0,1,2*PI);
}

int count_angle(float a1,float range){
	float x=world->player.PositionX;
	float y=world->player.PositionY;
	int c=0;
	
	forn(i,world->MAXMONSTERS){
		if(world->monsters[i].ClassName==-1)
			continue;
		if(in_angle(i,a1,range))
			++c;
	}
	return c;
}

const int safe_count=5;
void shoot_smart_forward_cloud(){
	int c=count_angle(atan2(world->player.RunY,world->player.RunX)-ang_range/2,ang_range);
	if(c<safe_count)
		shoot_cloud(0.2);
	else
		shoot_forward(0.3);
}

void shoot3(){
	forn(i,world->MAXMONSTERS)
		if(world->monsters[i].ClassName!=-1&&world->monsters[i].BulletVelocity>0){
			sniper_shot(i);
			return;
		}
	aim_shoot(0,1,2*PI);
}

const float round_prior = 100000000;
void move_round(float rad){
	float x=world->player.PositionX-0.5;
	float y=world->player.PositionY-0.5;
	float s1;
	if(x*x+y*y<rad*rad-0.05)
		s1=round_prior;
	else if(x*x+y*y>rad*rad+0.05)
		s1=-round_prior;
	else
		s1=0;
	float d=sqrt(x*x+y*y);
	if(d==0)
		d=1;
	x/=d;
	y/=d;
	float dx=x*s1+y;
	float dy=y*s1-x;
	d=sqrt(dx*dx+dy*dy);
	if(d==0)
		dx=1;
	else
		dx/=d,dy/=d;
	dx*=world->player.MaxVelocity;
	dy*=world->player.MaxVelocity;
	world->run(dx,dy);
}
void move_from(float fx,float fy){
	float x=world->player.PositionX;
	float y=world->player.PositionY;
	float dx=x-fx;
	float dy=y-fy;
	float d=sqrt(dx*dx+dy*dy);
	if(d==0)
		dx=1;
	else
		dx/=d,dy/=d;
	dx*=world->player.MaxVelocity;
	dy*=world->player.MaxVelocity;
	world->run(dx,dy);
}
void move_to(float fx,float fy){
	float x=world->player.PositionX;
	float y=world->player.PositionY;
	float dx=fx-x;
	float dy=fy-y;
	float d=sqrt(dx*dx+dy*dy);
	if(d==0)
		dx=1;
	else
		dx/=d,dy/=d;
	dx*=world->player.MaxVelocity;
	dy*=world->player.MaxVelocity;
	world->run(dx,dy);
}

void doit(){
	if(mode==SNIPER){
		move_round(0);
		sniper_shot();
	}else if(mode==SURVIVE){
		move_round(0.4);
		shoot_smart_forward_cloud();
	}else if(mode==ESCAPE){
		move_from(escape_x,escape_y);
		shoot_smart_forward_cloud();
	}else if(mode==CLOUD){
		move_round(0.4);
		shoot_cloud(0.2);
	}else if(mode==UPDOWN){
		if(world->player.PositionX<0.5-0.01)
			world->run(1000000,0);
		else if(world->player.PositionX>0.5+0.01)
			world->run(-1000000,0);
		else{
			static int d=0;
			if(d==0){
				if(world->player.PositionY<0.5)
					d=1;
				else
					d=-1;
			}
			//if(world->player.PositionY<0.01&&d==-1)
			//	d=1;
			//else if(world->player.PositionY>1-0.01&&d==1)
			//	d=-1;
			world->run(0,d*100000);
		}
		shoot3();
	}
}
const float across_margin=0.05;
const float good_rad=0.3;
void decide_mode(){
	static bool was2=false;
	if(wave2()){
		was2=true;
		if(monsters()>50)
			mode=UPDOWN;
		else
			mode=SNIPER;
		return;
	}
	if(was2&&monsters()>50){
		mode=UPDOWN;
		return;
	}
	float x=world->player.PositionX;
	float y=world->player.PositionY;
	float sx,sy,rad;
	get_circle(sx,sy,rad);
	if(SQ(sx-x)+SQ(sy-y)<SQ(rad)){
		if(mode!=SURVIVE&&min(min(x,1-x),min(y,1-y))<0.2)
				mode=SURVIVE;
		if(mode!=SURVIVE){
			mode=ESCAPE;
			escape_x=sx;
			escape_y=sy;
		}
	}else{
		mode=CLOUD;
		cloud_x=sx;
		cloud_y=sy;
	}
}

int main(int argc, char** argv) {
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	world = new World(argv[1], atoi(argv[2]));

	while (true) {
		world->makeMove();
		upd_shots();
		//forn(i,world->MAXMONSTERS)
		//	if(!can_shoot_at(i))
		//		cout<<i<<" is forbidden"<<endl;
		decide_mode();
		doit();
		debug_out();
	}

	return 0;
}
