#include "world.h"
#include <cstdio>
#include <cstring>
#include <cmath>
#include <algorithm>

using namespace std;

double const INF = 1e9;
double const EPS = 1e-9;
double const saveTime = 10;
const double magicR = 0.4;
World *world;


struct Point
{
	double x, y;
	Point(double x = 0, double y = 0): x(x), y(y) {};
	Point(CPlayer pl): x(pl.PositionX), y(pl.PositionY) {};
	Point(CMonster pl): x(pl.PositionX), y(pl.PositionY) {};
	double operator%(const Point & p) const
	{
		return x * p.x + y * p.y;
	}
	double len()
	{
		return sqrt(*this % *this);
	}
	Point operator-(const Point & p) const
	{
		return Point(x - p.x, y - p.y);
	}
	double dist(const Point & p) const
	{
		return (*this - p).len();
	}
};

double distance (CMonster monster)
{
	return sqrt ((world->player.PositionX - monster.PositionX) * (world->player.PositionX - monster.PositionX) +
				 (world->player.PositionY - monster.PositionY) * (world->player.PositionY - monster.PositionY));
}
double time (CMonster monster)
{
	if (monster.MaxVelocity < EPS)
		return INF;
	return distance (monster) / monster.MaxVelocity;
}

void run(double x, double y)
{
	const double c = 1.5;
	world->run(x / c, y / c);
}

int main(int argc, char** argv) 
{
	//Check for enough parameters
	if (argc < 3) {
		fprintf(stderr, "Parameters: IP port\n");
		return 10;
	}

	//Create the world data (and establish connection with the server)
	world = new World(argv[1], atoi(argv[2]));
	//float alpha = 0.0;

	//Infinite loop

	double avx = 0;
	double avy = 0;
	double maxR = 0;
	double w = 0;
	
	int ind = 0;
	while (true) 
	{
		//Send our turn and receive the update of the game world
		world->makeMove();

		//Print some information
		w += 0.01;
		if(ind++ % 2 == 0)
		{
			int cnt = 0;
			for(int i = 0; i < world->MAXMONSTERS; i++)
				if(world->monsters[i].ClassName >= 0 && distance(world->monsters[i]) < magicR)
				{
					cnt++;
					avx += world->monsters[i].PositionX;
					avy += world->monsters[i].PositionY;
				}
			maxR = 0;
			if(cnt > 0)
			{
				avx /= cnt;
				avy /= cnt;
				Point center = Point(avx, avy);
				for(int i = 0; i < world->MAXMONSTERS; i++)
				if(world->monsters[i].ClassName >= 0 && distance(world->monsters[i]) < magicR)
				{
					maxR = max(maxR, Point(world->monsters[i]).dist(center));
				}
			}
			maxR *= 100;
		}
		Point player = Point(world->player);
		Point center = Point(avx, avy);
		double curDist = player.dist(center);
		/*if(curDist < maxR)
		{
			Point dir = player - center;
			world->run(dir.x, dir.y);
		}
		else*/
		{
			fprintf(stdout, "maxR: %lf, avx: %lf, avy: %lf, w: %lf\n", maxR, avx, avy, w);
			
			Point dest(avx + maxR*cos(w), avy + maxR*sin(w));
			Point dir = dest - player;
			run(dir.x, dir.y);
		}

		double mh = INF;
		double mx = 0, my = 0;
		int chosen = -1;
		for(int i = 0; i < world->MAXMONSTERS; i++)
			if(world->monsters[i].ClassName >= 0 && (chosen == -1 || 
				(distance(world->monsters[i]) < distance(world->monsters[chosen]) || (fabs((distance(world->monsters[i]) - distance(world->monsters[chosen])) < 1e-9) && (world->monsters[i].HitPoints < mh)))))
			{
				mh = world->monsters[i].HitPoints;
				mx = world->monsters[i].PositionX;
				my = world->monsters[i].PositionY;
				chosen = i;
			}
			world->shoot(mx - player.x, my - player.y);
		//alpha += 0.02;
		//Run in circles (set direction)
		//world->run(cos(alpha), sin(alpha));

		//Choose the first monster that is still alive
		/*for ( int j = 0; j < world->MAXMONSTERS; j++)
			if (world->monsters[j].ClassName >= 0) {
				//Shoot the bastard! (Shoot in the direction of the monster)
				world->shoot(world->monsters[j].PositionX - world->player.PositionX,
						   world->monsters[j].PositionY - world->player.PositionY);
				break;
			}
			*/
	}

	return 0;
}
