{$H+}
unit world;
interface
uses simpleclass;

const WORLD_BEGIN_MARKER : longint  = 1243635;
const WORLD_END_MARKER : longint  = 5365456;
const CLIENT_BEGIN_MARKER : longint  = 42345; 
const CLIENT_END_MARKER : longint  = 8752845;
const MAXMONSTERS = 256;
const MAXMISSILES = 64;

Type TWorld = record 
	player : TCPlayer;
	monsters : array[1 .. MAXMONSTERS] of TCMonster;
	missiles: array[1 .. MAXMISSILES] of TCMissile;
end;

procedure makeMove(var m : TWorld);
procedure run(var m : TWorld; _vx : Single; _vy : Single);
procedure shoot(var m : TWorld;  _shootx : Single; _shooty : Single);

procedure Open(ip : string; port : longint);

implementation
uses CTypes, socketWrapper;

procedure makeMove(var m : TWorld);
var i, marker : longint;
begin
	lockMutex();

	writeInt(CLIENT_BEGIN_MARKER);
	WriteCPlayerToSocket(m.player);
	writeInt(CLIENT_END_MARKER);
	flush();



	marker := nextInt();
	if ( marker <> WORLD_BEGIN_MARKER) then
	begin
		writeln('Protocol error. Wrong WORLD_BEGIN_MARKER, expected ', WORLD_BEGIN_MARKER, ', got ', marker);
		halt(1);
	end;		

	ReadCPlayerFromSocket(m.player);
	for i := 1 to MAXMONSTERS do
		ReadCMonsterFromSocket(m.monsters[i]);

	for i := 1 to MAXMISSILES do
		ReadCMissileFromSocket(m.missiles[i]);

	marker := nextInt();
	if ( marker <> WORLD_END_MARKER ) then
	begin
		writeln('Protocol error. Wrong WORLD_END_MARKER, expected ', WORLD_END_MARKER, ', got ', marker);
		halt(1);
	end;		

	if ( m.player.HitPoints < 0.0 ) then
	begin
		writeln('Game over!');
		closeSocket();
		halt(0);
	end;

	unlockMutex();
end;

procedure run(var m : TWorld; _vx : Single; _vy : Single);
begin
	m.player.runX := _vx;
	m.player.runY := _vy;
end;

procedure shoot(var m : TWorld; _shootx : Single; _shooty : Single);
begin
	m.player.shootX := _shootx;
	m.player.shootY:= _shooty;
end;

procedure Open(ip : string; port : longint);
begin
	ConnectServer(PChar(ip), ctypes.cint32(port));
end;

end.
